#!/usr/bin/env python3

# Original source: https://github.com/chenxiaolong/DualBootPatcher

# Python 2 compatibility
from __future__ import print_function

import hashlib
import os
import struct
import sys
import traceback

BOOT_MAGIC = "ANDROID!"
BOOT_MAGIC_SIZE = 8
BOOT_NAME_SIZE = 16
BOOT_ARGS_SIZE = 512

MODE_STANDARD = 1
MODE_DEGAS = 2

def write_padding(f, pagesize, itemsize):
    pagemask = pagesize - 1

    if (itemsize & pagemask) == 0:
        return 0

    count = pagesize - (itemsize & pagemask)

    if sys.hexversion < 0x03000000:
        f.write(struct.pack('<' + str(count) + 's', '\0' * 131072))
    else:
        f.write(struct.pack('<' + str(count) + 's', bytes(0x00) * 131072))


show_output = True
use_stdout = False


def print_i(line):
    if show_output:
        if use_stdout:
            print(line)
        else:
            print(line, file=sys.stderr)


def build(filename, board=None, base=None, cmdline=None, page_size=None,
          kernel_offset=None, ramdisk_offset=None, second_offset=None,
          tags_offset=None, kernel=None, ramdisk=None, second=None, dt=None, signature = None, mode = MODE_STANDARD):
    # Defaults
    if board is None:          board          = ""
    if cmdline is None:        cmdline        = ""
    if base is None:           base           = 0x10000000
    if kernel_offset is None:  kernel_offset  = 0x00008000
    if ramdisk_offset is None: ramdisk_offset = 0x01000000
    if second_offset is None:  second_offset  = 0x00f00000
    if tags_offset is None:    tags_offset    = 0x00000100

    kernel_addr  = base + kernel_offset
    ramdisk_addr = base + ramdisk_offset
    second_addr  = base + second_offset
    tags_addr    = base + tags_offset

    if len(board) > BOOT_NAME_SIZE:
        raise Exception("Board name too large")

    if len(cmdline) > BOOT_ARGS_SIZE - 1:
        raise Exception("Kernel command line too large")

    kernel_data = None
    kernel_size = 0
    f = open(kernel, 'rb')
    kernel_data = f.read()
    f.close()
    kernel_size = len(kernel_data)

    ramdisk_data = None
    ramdisk_size = 0

    f = open(ramdisk, 'rb')
    ramdisk_data = f.read()
    f.close()
    ramdisk_size = len(ramdisk_data)

    second_data = None
    second_size = 0
    if second and os.path.exists(second):
        f = open(second, 'rb')
        second_data = f.read()
        f.close()
        second_size = len(second_data)

    dt_data = None
    dt_size = 0
    if dt and os.path.exists(dt):
        f = open(dt, 'rb')
        dt_data = f.read()
        f.close()
        dt_size = len(dt_data)

    sig_data = None
    sig_size = 0
    if signature and os.path.exists(signature):
        f = open(signature, 'rb')
        sig_data = f.read()
        f.close()
        sig_size = len(sig_data)

    sha = hashlib.sha1()
    sha.update(kernel_data)
    sha.update(struct.pack('<I', kernel_size))
    sha.update(ramdisk_data)
    sha.update(struct.pack('<I', ramdisk_size))
    if second_data:
        sha.update(second_data)
        sha.update(struct.pack('<I', second_size))
    if dt_data:
        sha.update(dt_data)
        sha.update(struct.pack('<I', dt_size))

    if sig_data:
        sha.update(sig_data)
        sha.update(struct.pack('<I', sig_size))

    f = open(filename, 'wb')

    sformat = '<'
    sformat += str(BOOT_MAGIC_SIZE) + 's'  # magic
    sformat += '10I'                       # kernel_size, kernel_addr,
                                           # ramdisk_size, ramdisk_addr,
                                           # second_size, second_addr,
                                           # tags_addr, page_size,
                                           # dt_size, unused
    sformat += str(BOOT_NAME_SIZE) + 's'   # name
    sformat += str(BOOT_ARGS_SIZE) + 's'   # cmdline
    sformat += str(8 * 4) + 's'            # id (unsigned[8])


    if mode == MODE_DEGAS:
        f.write(struct.pack(
            sformat,
            BOOT_MAGIC.encode('ascii'), # magic
            kernel_size,                # kernel_size
            kernel_addr,                # kernel_addr
            ramdisk_size,               # ramdisk_size
            ramdisk_addr,               # ramdisk_addr
            second_size,                # second_size
            second_addr,                # second_addr
            dt_size,                    # dt_size
            0x02000000,                          # unused
            tags_addr,                  # tags_addr
            page_size,                  # page_size
            board.encode('ascii'),      # name
            cmdline.encode('ascii'),    # cmdline
            sha.digest()                # id
        ))
    else:
        f.write(struct.pack(
            sformat,
            BOOT_MAGIC.encode('ascii'), # magic
            kernel_size,                # kernel_size
            kernel_addr,                # kernel_addr
            ramdisk_size,               # ramdisk_size
            ramdisk_addr,               # ramdisk_addr
            second_size,                # second_size
            second_addr,                # second_addr
            tags_addr,                  # tags_addr
            page_size,                  # page_size
            dt_size,                    # dt_size
            0,                          # unused
            board.encode('ascii'),      # name
            cmdline.encode('ascii'),    # cmdline
            sha.digest()                # id
        ))

    write_padding(f, page_size, struct.calcsize(sformat))

    f.write(kernel_data)
    write_padding(f, page_size, len(kernel_data))

    f.write(ramdisk_data)
    write_padding(f, page_size, len(ramdisk_data))

    if second_data:
        f.write(second_data)
        write_padding(f, page_size, len(second_data))

    if dt_data:
        f.write(dt_data)
        write_padding(f, page_size, len(dt_data))

    if sig_data:
        f.write(sig_data)
        write_padding(f, page_size, len(sig_data))

    f.close()


def auto_from_dir(directory, prefix, mode = MODE_STANDARD):
    prefix = os.path.join(directory, prefix)

    base = None
    if os.path.exists(prefix + '-base'):
        f = open(prefix + '-base', 'rb')
        base = int(f.readline().decode('ASCII').rstrip('\n'), 16)
        f.close()

    cmdline = None
    if os.path.exists(prefix + '-cmdline'):
        f = open(prefix + '-cmdline', 'rb')
        cmdline = f.readline().decode('ASCII').rstrip('\n')
        f.close()

    page_size = None
    if os.path.exists(prefix + '-pagesize'):
        f = open(prefix + '-pagesize', 'rb')
        page_size = int(f.readline().decode('ASCII').rstrip('\n'))
        f.close()

    ramdisk_offset = None
    if os.path.exists(prefix + '-ramdisk_offset'):
        f = open(prefix + '-ramdisk_offset', 'rb')
        ramdisk_offset = int(f.readline().decode('ASCII').rstrip('\n'), 16)
        f.close()

    second_offset = None
    if os.path.exists(prefix + '-second_offset'):
        f = open(prefix + '-second_offset', 'rb')
        second_offset = int(f.readline().decode('ASCII').rstrip('\n'), 16)
        f.close()

    tags_offset = None
    if os.path.exists(prefix + '-tags_offset'):
        f = open(prefix + '-tags_offset', 'rb')
        tags_offset = int(f.readline().decode('ASCII').rstrip('\n'), 16)
        f.close()

    kernel = None
    if os.path.exists(prefix + '-zImage'):
        kernel = prefix + '-zImage'

    ramdisk = None
    if os.path.exists(prefix + '-ramdisk.gz'):
        ramdisk = prefix + '-ramdisk.gz'
    elif os.path.exists(prefix + '-ramdisk.lz4'):
        ramdisk = prefix + '-ramdisk.lz4'

    second = None
    if os.path.exists(prefix + '-second'):
        second = prefix + '-second'

    dt = None
    if os.path.exists(prefix + '-dt'):
        dt = prefix + '-dt'

    signature = None
    if os.path.exists(prefix + '-signature'):
        signature = prefix + '-signature'

    build(output,
          board=None,
          base=base,
          cmdline=cmdline,
          page_size=page_size,
          kernel_offset=None,
          ramdisk_offset=ramdisk_offset,
          second_offset=second_offset,
          tags_offset=tags_offset,
          kernel=kernel,
          ramdisk=ramdisk,
          second=second,
          dt=dt,
          signature=signature,
          mode=mode)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print_i("Usage: %s" % sys.argv[0])
        print_i("       -o|--output <filename>")
        print_i("")
        print_i("Manual mode:")
        print_i("       [ --board <name> ]")
        print_i("       [ --base <address> ]")
        print_i("       [ --cmdline <kernel-cmdline> ]")
        print_i("       [ --pagesize <size> ]")
        print_i("       [ --kernel_offset <address> ]")
        print_i("       [ --ramdisk_offset <address> ]")
        print_i("       [ --second_offset <address> ]")
        print_i("       [ --tags_offset <address> ]")
        print_i("       --kernel <filename>")
        print_i("       --ramdisk <filename>")
        print_i("       [ --second <filename> ]")
        print_i("       [ --dt <filename> ]")
        print_i("       [ --signature <filename> ]")
        print_i("       [ --degas ]")
        print_i("")
        print_i("Automatic mode:")
        print_i("       --auto_dir <directory>")
        print_i("       --auto_prefix <prefix>")
        sys.exit(1)

    board = None
    base = None
    cmdline = None
    page_size = None
    kernel_offset = None
    ramdisk_offset = None
    second_offset = None
    tags_offset = None
    kernel = None
    ramdisk = None
    second = None
    dt = None
    output = None
    signature = None
    degas = False

    auto_dir = None
    auto_base = None

    i = 1
    while i < len(sys.argv):
        if sys.argv[i] == '--board':
            board = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--base':
            base = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--cmdline':
            cmdline = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--pagesize':
            page_size = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--kernel_offset':
            kernel_offset = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--ramdisk_offset':
            ramdisk_offset = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--second_offset':
            second_offset = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--tags_offset':
            tags_offset = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--kernel':
            kernel = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--ramdisk':
            ramdisk = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--second':
            second = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--dt':
            dt = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '-o' or sys.argv[i] == '--output':
            output = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--auto_dir':
            auto_dir = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--auto_prefix':
            auto_base = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--signature':
            signature = sys.argv[i + 1]
            i += 2
        elif sys.argv[i] == '--degas':
            degas = True
            i += 1
        else:
            print_i("Unrecognized argument " + sys.argv[i])
            sys.exit(1)

    if not output:
        print_i("No output filename specified")
        sys.exit(1)

    if (auto_dir and not auto_base) or (auto_base and not auto_dir):
        print_i("Automatic mode used, but second argument is missing")
        sys.exit(1)

    try:
        use_stdout = True

        if auto_dir and auto_base:
            auto_from_dir(auto_dir, auto_base, mode=MODE_DEGAS if degas else MODE_STANDARD)

        else:
            if not kernel:
                print_i("No kernel specified")
                sys.exit(1)

            if not ramdisk:
                print_i("No ramdisk specified")
                sys.exit(1)

            if base:           base           = int(base, 16)
            if page_size:      page_size      = int(page_size)
            if kernel_offset:  kernel_offset  = int(kernel_offset, 16)
            if ramdisk_offset: ramdisk_offset = int(ramdisk_offset, 16)
            if second_offset:  second_offset  = int(second_offset, 16)
            if tags_offset:    tags_offset    = int(tags_offset, 16)

            build(output,
                  board=board,
                  base=base,
                  cmdline=cmdline,
                  page_size=page_size,
                  kernel_offset=kernel_offset,
                  ramdisk_offset=ramdisk_offset,
                  second_offset=second_offset,
                  tags_offset=tags_offset,
                  kernel=kernel,
                  ramdisk=ramdisk,
                  second=second,
                  dt=dt,
                  signature=signature,
                  mode=MODE_DEGAS if degas else MODE_STANDARD
                  )
    except Exception as e:
        use_stdout = False
        print_i("Failed: " + str(e))
        print(traceback.format_exc())


#!/usr/bin/env python3
# Original source: https://github.com/chenxiaolong/DualBootPatcher
# Python 2 compatibility
from __future__ import print_function

import binascii
import os
import struct
import sys
import traceback
from bootimg import  BootImg

BOOT_MAGIC = "ANDROID!"
BOOT_MAGIC_SIZE = 8
BOOT_NAME_SIZE = 16
BOOT_ARGS_SIZE = 512

MODE_STANDARD = 1
MODE_DEGAS = 2


def read_padding(f, itemsize, pagesize):
    pagemask = pagesize - 1

    if (itemsize & pagemask) == 0:
        return 0

    count = pagesize - (itemsize & pagemask)

    f.read(count)

    return count


def bytes_to_str(data):
    temp = binascii.hexlify(data).decode("utf-8")
    #return "\\x" + "\\x".join(a + b for a, b in zip(temp[::2], temp[1::2]))
    return "".join(a + b for a, b in zip(temp[::2], temp[1::2]))


show_output = True
use_stdout = False


def print_i(line):
    if show_output:
        if use_stdout:
            print(line)
        else:
            print(line, file=sys.stderr)


def extract(filename, directory, mode = MODE_STANDARD):
    basename = os.path.split(filename)[1]
    f = open(filename, 'rb')

    i = 0
    while i <= 512:
        f.seek(i, os.SEEK_SET)
        temp = f.read(BOOT_MAGIC_SIZE)
        if temp == BOOT_MAGIC.encode("ASCII"):
            # print_i("Found Android header")
            break
        i += 1

    if i > 512:
        raise Exception("Android header not found")

    # Read Android header
    f.seek(0, os.SEEK_SET)

    sformat = '<'
    sformat += str(BOOT_MAGIC_SIZE) + 's'  # magic
    sformat += '10I'                       # kernel_size, kernel_addr,
                                           # ramdisk_size, ramdisk_addr,
                                           # second_size, second_addr,
                                           # tags_addr, page_size,
                                           # dt_size, unused
    sformat += str(BOOT_NAME_SIZE) + 's'   # name
    sformat += str(BOOT_ARGS_SIZE) + 's'   # cmdline
    sformat += str(8 * 4) + 's'            # id (unsigned[8])

    header_size = struct.calcsize(sformat)
    header = f.read(header_size)

    if mode == MODE_DEGAS:
        magic, kernel_size, kernel_addr, \
            ramdisk_size, ramdisk_addr, \
            second_size, second_addr, \
            dt_size, unused, \
            tags_addr, page_size, \
            board, cmdline, ident = struct.unpack(sformat, header)
    else:
        magic, kernel_size, kernel_addr, \
            ramdisk_size, ramdisk_addr, \
            second_size, second_addr, \
            tags_addr, page_size, \
            dt_size, unused, \
            board, cmdline, ident = struct.unpack(sformat, header)

    ramdisk_offset = ramdisk_addr - kernel_addr + 0x00008000
    tags_offset = tags_addr - kernel_addr + 0x00008000
    second_offset = second_addr - kernel_addr + 0x00008000
    base = kernel_addr - 0x00008000
    kernel_offset = kernel_addr - base
    kernel = os.path.join(directory, basename + "-zImage")
    dt = os.path.join(directory, basename + "-dt")
    cmdline =  cmdline.decode('ASCII').rstrip('\0')
    signature = os.path.join(directory, basename + "-signature")

    # cmdline
    out = open(os.path.join(directory, basename + "-cmdline"), 'wb')
    out.write((cmdline + '\n').encode('ASCII'))
    out.close()

    # base
    out = open(os.path.join(directory, basename + "-base"), 'wb')
    out.write(('%08x\n' % (kernel_addr - 0x00008000)).encode('ASCII'))
    out.close()

    # ramdisk_offset
    out = open(os.path.join(directory, basename + "-ramdisk_offset"), 'wb')
    out.write(('%08x\n' % ramdisk_offset).encode('ASCII'))
    out.close()

    # second_offset
    out = open(os.path.join(directory, basename + "-second_offset"), 'wb')
    out.write(('%08x\n' % second_offset).encode('ASCII'))
    out.close()

    # tags_offset
    out = open(os.path.join(directory, basename + "-tags_offset"), 'wb')
    out.write(('%08x\n' % tags_offset).encode('ASCII'))
    out.close()

    # page_size
    out = open(os.path.join(directory, basename + "-pagesize"), 'wb')
    out.write(('%d\n' % page_size).encode('ASCII'))
    out.close()

    read_padding(f, header_size, page_size)

    # zImage
    out = open(kernel, 'wb')
    k = f.read(kernel_size)
    out.write(k)
    out.close()

    read_padding(f, kernel_size, page_size)

    # ramdisk
    ramdisk = f.read(ramdisk_size)
    if ramdisk[0] == 0x02 and ramdisk[1] == 0x21:
        ramdisk_path = os.path.join(directory, basename + "-ramdisk.lz4")
    else:
        ramdisk_path = os.path.join(directory, basename + "-ramdisk.gz")

    out = open(ramdisk_path, 'wb')
    out.write(ramdisk)
    out.close()

    read_padding(f, ramdisk_size, page_size)

    # second
    out = open(os.path.join(directory, basename + "-second"), 'wb')
    out.write(f.read(second_size))
    out.close()

    read_padding(f, second_size, page_size)

    # dt
    out = open(dt, 'wb')
    out.write(f.read(dt_size))
    out.close()





    img = BootImg(
        board = board,
        base=base,
        cmdline=cmdline,
        page_size=page_size,
        kernel_offset=kernel_offset,
        ramdisk_offset=ramdisk_offset,
        second_offset=second_offset,
        tags_offset=tags_offset,
        kernel=kernel,
        ramdisk=ramdisk_path,
        second=None,
        dt=dt
    )

    if mode == MODE_DEGAS:
        #signature
        out = open(signature, 'wb')
        out.write(f.read(256))
        out.close()
        img.signature = signature

    f.close()

    print(img)
    return img


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print_i("Usage: %s -i boot.img [-o output_directory]" % sys.argv[0])
        sys.exit(1)

    filename = ""
    directory = ""

    counter = 1
    degas = False
    while counter < len(sys.argv):
        if sys.argv[counter] == "-i":
            filename = sys.argv[counter + 1]
            counter += 2
        elif sys.argv[counter] == "-o":
            directory = sys.argv[counter + 1]
            counter += 2
        elif sys.argv[counter] == '--degas':
            degas = True
            counter += 1
        else:
            print_i("Unrecognized argument " + sys.argv[counter])
            sys.exit(1)

    if filename == "":
        print_i("No boot image specified")
        sys.exit(1)

    if not os.path.exists(filename):
        print_i(filename + " does not exist")
        sys.exit(1)

    if directory == "":
        directory = os.getcwd()
    elif not os.path.exists(directory):
        os.makedirs(directory)

    try:
        use_stdout = True
        extract(filename, directory, MODE_DEGAS if degas else MODE_STANDARD)
    except Exception as e:
        use_stdout = False
        print(traceback.format_exc())
        print_i("Failed: " + str(e))


#!/usr/bin/env python3
# Original source: https://github.com/chenxiaolong/DualBootPatcher
# Python 2 compatibility
from __future__ import print_function

import binascii
import gzip
import os
import struct
import sys
import zlib

BOOT_MAGIC = "ANDROID!"
BOOT_MAGIC_SIZE = 8
BOOT_NAME_SIZE = 16
BOOT_ARGS_SIZE = 512


def bytes_to_str(data):
    temp = binascii.hexlify(data).decode("utf-8")
    #return "\\x" + "\\x".join(a + b for a, b in zip(temp[::2], temp[1::2]))
    return "".join(a + b for a, b in zip(temp[::2], temp[1::2]))


show_output = True
use_stdout = False


def print_i(line):
    if show_output:
        if use_stdout:
            print(line)
        else:
            print(line, file=sys.stderr)


def extract(filename, directory):
    basename = os.path.split(filename)[1]
    f = open(filename, 'rb')

    i = 0
    while i <= 32:
        f.seek(i, os.SEEK_SET)
        temp = f.read(BOOT_MAGIC_SIZE)
        if temp == BOOT_MAGIC.encode("ASCII"):
            print_i("Found Android header")
            break
        i += 1

    if i > 512:
        raise Exception("Android header not found")

    # Read Android header
    f.seek(0, os.SEEK_SET)

    sformat = '<'
    sformat += str(BOOT_MAGIC_SIZE) + 's'  # magic
    sformat += '10I'                       # kernel_size, kernel_addr,
                                           # ramdisk_size, ramdisk_addr,
                                           # second_size, second_addr,
                                           # tags_addr, page_size,
                                           # dt_size, unused
    sformat += str(BOOT_NAME_SIZE) + 's'   # name
    sformat += str(BOOT_ARGS_SIZE) + 's'   # cmdline
    sformat += str(8 * 4) + 's'            # id (unsigned[8])

    header_size = struct.calcsize(sformat)
    header = f.read(header_size)

    magic, kernel_size, kernel_addr, \
        ramdisk_size, ramdisk_addr, \
        second_size, second_addr, \
        tags_addr, page_size, \
        dt_size, unused, \
        board, cmdline, ident = struct.unpack(sformat, header)

    print_i("- magic:        " + magic.decode('ASCII'))
    print_i("- kernel_size:  " + str(kernel_size))
    print_i("- kernel_addr:  " + hex(kernel_addr))
    print_i("- ramdisk_size: " + str(ramdisk_size))
    print_i("- ramdisk_addr: " + hex(ramdisk_addr))
    print_i("- second_size:  " + str(second_size))
    print_i("- second_addr:  " + hex(second_addr))
    print_i("- tags_addr:    " + hex(tags_addr))
    print_i("- page_size:    " + str(page_size))
    print_i("- dt_size:      " + str(dt_size))
    print_i("- unused:       " + hex(unused))
    print_i("- name:         " + board.decode('ASCII'))
    print_i("- cmdline:      " + cmdline.decode('ASCII'))
    print_i("- id:           " + bytes_to_str(ident)[:40])  # SHA1 is 40 chars

    print_i("")

    # Look for loki header
    f.seek(0x400, os.SEEK_SET)

    lokformat = '<'
    lokformat += '4s'    # magic (0x494b4f4c)
    lokformat += 'I'     # recovery (0 = boot.img, 1 = recovery.img)
    lokformat += '128s'  # build number
    lokformat += '3I'    # orig_kernel_size, orig_ramdisk_size, ramdisk_addr

    lok_header_size = struct.calcsize(lokformat)
    lok_header = f.read(lok_header_size)

    lok_magic, lok_recovery, lok_build, \
        lok_orig_kernel_size, lok_orig_ramdisk_size, \
        lok_ramdisk_addr, = struct.unpack(lokformat, lok_header)

    if lok_magic == "LOKI".encode("ASCII"):
        print_i("Found Loki header")
    else:
        raise Exception("Could not find Loki header")

    print_i("- magic:             " + lok_magic.decode('ascii'))
    print_i("- recovery:          " + str(lok_recovery))
    print_i("- build:             " + lok_build.decode('ascii'))
    print_i("- orig_kernel_size:  " + str(lok_orig_kernel_size))
    print_i("- orig_ramdisk_size: " + str(lok_orig_ramdisk_size))
    print_i("- ramdisk_addr:      " + hex(lok_ramdisk_addr))

    print_i("")

    # Get total size
    f.seek(0, os.SEEK_END)
    total_size = f.tell()

    # Find ramdisk's gzip header
    gzip_offset = 0x400 + 148  # size of Loki header
    while gzip_offset < total_size:
        f.seek(gzip_offset, os.SEEK_SET)
        temp = f.read(4)
        timestamp = f.read(4)
        # Searching for 1F8B0800 wasn't enough for some boot images. Specifically,
        # ktoonsez's 20140319 kernels had another set of those four bytes before
        # the "real" gzip header. We'll work around that by checking that the
        # timestamp isn't zero (which is technically allowed, but the boot image
        # tools don't do that)
        # http://forum.xda-developers.com/showpost.php?p=51219628&postcount=3767
        if temp == b"\x1F\x8B\x08\x00" and timestamp != b"\x00\x00\x00\x00":
            print_i("Found gzip header at: " + hex(gzip_offset))
            break
        gzip_offset += 4

    # The ramdisk is supposed to be from the gzip header to EOF, but loki needs
    # to store a copy of aboot, so it is stored in 0x200 bytes are the end,
    # which Dan Rosenberg calls the "fake" area.
    gzip_size = total_size - gzip_offset - 0x200
    # The line above is for Samsung kernels only. Uncomment this one for use on
    # LG kernels:
    #gzip_size = total_size - gzip_offset - page_size
    print_i("Ramdisk size: %i (may include some padding)" % gzip_size)

    # Get kernel size
    # Unfortunately, the original size is not stored in the Loki header on
    # kernels for Samsung devices and we can't reverse Dan Rosenberg's
    # expression because data is loss with the bitwise and:
    # hdr->kernel_size = ((hdr->kernel_size + page_mask) & ~page_mask) + hdr->ramdisk_size;
    # Luckily, the zimage headers stores the size at 0x2c, so we can get it
    # from there.
    # http://www.simtec.co.uk/products/SWLINUX/files/booting_article.html#d0e309
    f.seek(page_size + 0x2c, os.SEEK_SET)
    kernel_size = struct.unpack('<I', f.read(4))[0]
    print_i("Kernel size: " + str(kernel_size))

    print_i("")

    print_i("Writing kernel command line to %s-cmdline ..." % basename)
    out = open(os.path.join(directory, basename + "-cmdline"), 'wb')
    out.write((cmdline.decode("ascii").rstrip('\0') + '\n').encode("ascii"))
    out.close()

    print_i("Writing base address to %s-base ..." % basename)
    out = open(os.path.join(directory, basename + "-base"), 'wb')
    out.write(('%08x\n' % (kernel_addr - 0x00008000)).encode('ASCII'))
    out.close()

    print_i("Writing page size to %s-pagesize ..." % basename)
    out = open(os.path.join(directory, basename + "-pagesize"), 'wb')
    out.write(('%d\n' % page_size).encode('ASCII'))
    out.close()

    print_i("Writing kernel to %s-zImage ..." % basename)
    f.seek(page_size, os.SEEK_SET)
    kernel_data = f.read(kernel_size)
    out = open(os.path.join(directory, basename + "-zImage"), 'wb')
    out.write(kernel_data)
    out.close()

    # Truncate padding by recompression. Due to a bug in Python's gzip module,
    # we can't use it
    # http://stackoverflow.com/questions/4928560/how-can-i-work-with-gzip-files-which-contain-extra-data
    print_i("Writing ramdisk to %s-ramdisk.gz ..." % basename)
    f.seek(gzip_offset, os.SEEK_SET)
    ramdisk_data = f.read(gzip_size)
    out = gzip.open(os.path.join(directory, basename + "-ramdisk.gz"), 'wb')
    out.write(zlib.decompress(ramdisk_data, 16 + zlib.MAX_WBITS))
    out.close()

    f.close()

    print_i("")
    print_i("Successfully unloki'd " + filename)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print_i("Usage: %s -i boot.img [-o output_directory]" % sys.argv[0])
        sys.exit(1)

    filename = ""
    directory = ""

    counter = 1
    while counter < len(sys.argv):
        if sys.argv[counter] == "-i":
            filename = sys.argv[counter + 1]
            counter += 2
        elif sys.argv[counter] == "-o":
            directory = sys.argv[counter + 1]
            counter += 2
        else:
            print_i("Unrecognized argument " + sys.argv[counter])
            sys.exit(1)

    if filename == "":
        print_i("No loki image specified")
        sys.exit(1)

    if not os.path.exists(filename):
        print_i(filename + " does not exist")
        sys.exit(1)

    if directory == "":
        directory = os.getcwd()
    elif not os.path.exists(directory):
        os.makedirs(directory)

    try:
        use_stdout = True
        extract(filename, directory)
    except Exception as e:
        use_stdout = False
        print_i("Failed: " + str(e))


import os
import sys
import mkbootimg

if sys.version_info >= (3,0):
    unicode = str

class BootImg(object):

    def __init__(self,
                 board=None, base=None, cmdline=None, page_size=None,
                 kernel_offset=None, ramdisk_offset=None, second_offset=None,
                 tags_offset=None, kernel=None, ramdisk=None, second=None, dt=None,
                 signature=None):

        self.board = board
        self.base = base
        self.cmdline = cmdline
        self.page_size = page_size
        self.kernel_offset = kernel_offset
        self.ramdisk_offset = ramdisk_offset
        self.second_offset = second_offset
        self.tags_offset = tags_offset
        self.kernel = kernel
        self.ramdisk = ramdisk
        self.second = second
        self.dt = dt
        self.signature = signature

    def __int_address(self, addr):
        if addr is None:
            return None

        return int(addr, 16) if type(addr) in (str, unicode) else addr

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, value):
        self._board = value

    @property
    def base(self):
        return self._base

    @base.setter
    def base(self, value):
        self._base = self.__int_address(value)

    @property
    def cmdline(self):
        return self._cmdline

    @cmdline.setter
    def cmdline(self, value):
        self._cmdline = value

    @property
    def page_size(self):
        return self._page_size

    @page_size.setter
    def page_size(self, value):
        self._page_size = int(value) if value is not None else None

    @property
    def kernel_offset(self):
        return self._kernel_offset

    @kernel_offset.setter
    def kernel_offset(self, value):
        self._kernel_offset = self.__int_address(value)

    @property
    def ramdisk_offset(self):
        return self._ramdisk_offset

    @ramdisk_offset.setter
    def ramdisk_offset(self, value):
        self._ramdisk_offset = self.__int_address(value)

    @property
    def second_offset(self):
        return self._second_offset

    @second_offset.setter
    def second_offset(self, value):
        self._second_offset = self.__int_address(value)

    @property
    def tags_offset(self):
        return self._tags_offset

    @tags_offset.setter
    def tags_offset(self, value):
        self._tags_offset = self.__int_address(value)

    @property
    def kernel(self):
        return self._kernel

    @kernel.setter
    def kernel(self, value):
        self._kernel = value

    @property
    def ramdisk(self):
        return self._ramdisk

    @ramdisk.setter
    def ramdisk(self, value):
        self._ramdisk = value

    @property
    def second(self):
        return self._second

    @second.setter
    def second(self, value):
        self._second = value

    @property
    def second_size(self):
        return os.path.getsize(self._second) if self._second and os.path.exists(self._second) else 0

    @property
    def dt(self):
        return self._dt

    @property
    def dt_size(self):
        return os.path.getsize(self._dt) if self._dt and os.path.exists(self._dt) else 0

    @dt.setter
    def dt(self, value):
        self._dt = value

    @property
    def signature(self):
        return self._signature

    @signature.setter
    def signature(self, value):
        self._signature = value

    def build(self, filename, mode=mkbootimg.MODE_STANDARD):
        import mkbootimg
        mkbootimg.build(
            filename,
            self.board,
            self.base,
            self.cmdline,
            self.page_size,
            self.kernel_offset,
            self.ramdisk_offset,
            self.second_offset,
            self.tags_offset,
            self.kernel,
            self.ramdisk,
            self.second,
            self.dt,
            self.signature
        )

    def __str__(self):
        out = "BOARD_KERNEL_CMDLINE %s\n" % self.cmdline
        out += "BOARD_KERNEL_BASE %08x\n" % self.base
        out += "BOARD_RAMDISK_OFFSET %08x\n" % self.ramdisk_offset
        out += "BOARD_SECOND_OFFSET %08x\n" % self.second_offset
        out += "BOARD_TAGS_OFFSET %08x\n" % self.tags_offset
        out += "BOARD_PAGE_SIZE %s\n" % self.page_size
        out += "BOARD_SECOND_SIZE %s\n" % self.second_size
        out += "BOARD_DT_SIZE %s" % self.dt_size

        return out


from __future__ import print_function
from glob import glob
from setuptools import setup, find_packages, Extension

sources = ['c/droidtools_ext4fsmodule.c']
sources.extend(glob("c/ext4_utils/*.c"))
module_ext4 = Extension('droidtools.ext4fs',
                    define_macros = [('ANDROID', None)],
                    sources = sources,
                    libraries = ['z'],
                    include_dirs = [
                      'c/ext4_utils',
                      'c/ext4_utils/private'
                      ]
                    )
setup(
    name='python-ext4fs',
    packages= find_packages(),
    version="0.1a14",
    license='GPLv3 License',
    author='Tarek Galal',
    ext_modules = [module_ext4],
    author_email='tare2.galal@gmail.com',
    description='Some android tools',
    platforms='any',
    classifiers =[
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# Original source: https://github.com/chenxiaolong/DualBootPatcher
#
# Copyright (c) 2014 Xiao-Long Chen <chenxiaolong@cxl.epac.to>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# This is a miniature implementation of cpio supporting only enough of the
# file format for the purpose of the dual boot patcher. Only the "new format"
# for the cpio entries are supported. Because of the use case, everything is
# loaded into memory (needs 2x the file size). Don't use this with large files!

# Supported files
# - Regular files
# - Directories
# - Symlinks
# - Hardlinks


import os
import stat
import struct

###
debug = False
###

MAGIC_NEW = "070701"                 # new format
MAGIC_NEW_CRC = "070702"             # new format w/crc
MAGIC_OLD_ASCII = "070707"           # old ascii format
MAGIC_OLD_BINARY = int("070707", 8)  # old binary format

# Constants from cpio.h

# A header with a filename "TRAILER!!!" indicates the end of the archive.
CPIO_TRAILER = "TRAILER!!!"

IO_BLOCK_SIZE = 512


class CpioEntryNew:
    def __init__(self):
        self.sformat = '<'
        self.sformat += '6s'  # c_magic     - "070701" for "new" portable format
                              #               "070702" for CRC format
        self.sformat += '8s'  # c_ino
        self.sformat += '8s'  # c_mode
        self.sformat += '8s'  # c_uid
        self.sformat += '8s'  # c_gid
        self.sformat += '8s'  # c_nlink
        self.sformat += '8s'  # c_mtime
        self.sformat += '8s'  # c_filesize  - must be 0 for FIFOs and directories
        self.sformat += '8s'  # c_dev_maj
        self.sformat += '8s'  # c_dev_min
        self.sformat += '8s'  # c_rdev_maj  - only valid for chr and blk special files
        self.sformat += '8s'  # c_rdev_min  - only valid for chr and blk special files
        self.sformat += '8s'  # c_namesize  - count includes terminating NUL in pathname
        self.sformat += '8s'  # c_chksum    - 0 for "new" portable format; for CRC format
                              #               the sum of all the bytes in the file

        self.headersize = 0
        self.alignmentBoundary = 0

        self.magic = None
        self.ino = None
        self.mode = None
        self.uid = None
        self.gid = None
        self.nlink = None
        self.mtime = None
        self.filesize = None
        self.dev_maj = None
        self.dev_min = None
        self.rdev_maj = None
        self.rdev_min = None
        self.namesize = None
        self.chksum = None

        self.name = None
        self.content = None

        self.headersize = 110
        self.alignmentBoundary = 4

    def unpack(self, data, offset=0):
        unpacked = struct.unpack_from(self.sformat, data, offset)

        self.magic = unpacked[0].decode('ascii')
        self.ino = int(unpacked[1], 16)
        self.mode = int(unpacked[2], 16)
        self.uid = int(unpacked[3], 16)
        self.gid = int(unpacked[4], 16)
        self.nlink = int(unpacked[5], 16)
        self.mtime = int(unpacked[6], 16)
        self.filesize = int(unpacked[7], 16)
        self.dev_maj = int(unpacked[8], 16)
        self.dev_min = int(unpacked[9], 16)
        self.rdev_maj = int(unpacked[10], 16)
        self.rdev_min = int(unpacked[11], 16)
        self.namesize = int(unpacked[12], 16)
        self.chksum = int(unpacked[13], 16)

        namestart = offset + struct.calcsize(self.sformat)
        nameend = namestart + self.namesize

        self.name = data[namestart:nameend - 1].decode('ascii')

        padding = (4 - (nameend % 4)) % 4

        datastart = nameend + padding
        dataend = datastart + self.filesize

        self.content = data[datastart:dataend]

        padding = (4 - (dataend % 4)) % 4

        return dataend + padding

    def pack(self, data):
        header = struct.pack(
            self.sformat,
            self.magic.encode('ascii'),
            ('%08x' % self.ino).encode('ascii'),
            ('%08x' % self.mode).encode('ascii'),
            ('%08x' % self.uid).encode('ascii'),
            ('%08x' % self.gid).encode('ascii'),
            ('%08x' % self.nlink).encode('ascii'),
            ('%08x' % self.mtime).encode('ascii'),
            ('%08x' % self.filesize).encode('ascii'),
            ('%08x' % self.dev_maj).encode('ascii'),
            ('%08x' % self.dev_min).encode('ascii'),
            ('%08x' % self.rdev_maj).encode('ascii'),
            ('%08x' % self.rdev_min).encode('ascii'),
            ('%08x' % self.namesize).encode('ascii'),
            ('%08x' % self.chksum).encode('ascii'))

        offset = len(data)

        data += header

        # name
        data += self.name.encode('ascii') + b'\x00'

        namestart = offset + struct.calcsize(self.sformat)
        nameend = namestart + self.namesize
        padding = (4 - (nameend % 4)) % 4

        data += b'\x00' * padding

        # content
        data += self.content

        datastart = nameend + padding
        dataend = datastart + self.filesize
        padding = (4 - (dataend % 4)) % 4

        data += b'\x00' * padding

        return data

    def set_content(self, content):
        self.content = content
        self.filesize = len(content)

    def dump(self):
        #C_ISCTG  = int("0110000", 8)
        filetype = stat.S_IFMT(self.mode)

        if stat.S_ISDIR(self.mode):
            ftypestr = "directory"
        elif stat.S_ISLNK(self.mode):
            ftypestr = "symbolic link"
        elif stat.S_ISREG(self.mode):
            ftypestr = "regular file"
        elif stat.S_ISFIFO(self.mode):
            ftypestr = "pipe"
        elif stat.S_ISCHR(self.mode):
            ftypestr = "character device"
        elif stat.S_ISBLK(self.mode):
            ftypestr = "block device"
        elif stat.S_ISSOCK(self.mode):
            ftypestr = "socket"
        #elif filetype == C_ISCTG:
        #    ftypestr = "what is this"
        else:
            ftypestr = 'unknown (%o)' % filetype

        print('Filename:        %s' % self.name)
        print('Filetype:        %s' % ftypestr)
        print('Magic:           %s' % self.magic)
        print('Inode:           %i' % self.ino)
        print('Mode:            %o' % self.mode)
        print('Permissions:     %o' % (self.mode - filetype))
        print('UID:             %i' % self.uid)
        print('GID:             %i' % self.gid)
        print('Links:           %i' % self.nlink)
        print('Modified:        %i' % self.mtime)
        print('File size:       %i' % self.filesize)
        print('dev_maj:         %x' % self.dev_maj)
        print('dev_min:         %x' % self.dev_min)
        print('rdev_maj:        %x' % self.rdev_maj)
        print('rdev_min:        %x' % self.rdev_min)
        print('Filename length: %i' % self.namesize)
        print('Checksum:        %i' % self.chksum)
        print('')


class CpioFile:
    def __init__(self):
        self.members = []
        self.inodemap = dict()

    def load_file(self, filename):
        with open(filename, 'rb') as f:
            self.load(f.read())

    def load(self, data):
        pos = 0

        while True:
            member = CpioEntryNew()
            pos = member.unpack(data, pos)

            if debug:
                member.dump()

            if member.name == CPIO_TRAILER:
                break

            # Make sure the existing inodes won't conflict with ones from newly
            # added files, so make them negative for now

            member.ino *= -1

            if member.ino in self.inodemap:
                self.inodemap[member.ino].append(member)
            else:
                self.inodemap[member.ino] = [ member ]

            self.members.append(member)

    def create(self):
        buf = bytes()

        # Uniquify inodes
        cur_inode = 300000

        # The kernel wants directories to come first
        temp = []

        # Assign inodes
        for inode in self.inodemap:
            for i in self.inodemap[inode]:
                i.ino = cur_inode
                temp.append(i)

            cur_inode += 1

        temp = sorted(temp, key=lambda x: x.name)

        for i in temp:
            buf = i.pack(buf)

        member = CpioEntryNew()
        member.magic = MAGIC_NEW
        member.ino = cur_inode
        member.mode = 0
        member.uid = 0
        member.gid = 0
        member.nlink = 1  # Must be 1 for crc format
        member.mtime = 0
        member.filesize = 0
        member.dev_maj = 0
        member.dev_min = 0
        member.rdev_maj = 0
        member.rdev_min = 0
        member.namesize = len(CPIO_TRAILER) + 1
        member.chksum = 0
        member.name = CPIO_TRAILER
        member.content = b''
        buf = member.pack(buf)

        # Pad until end of block
        buf += b'\x00' * (IO_BLOCK_SIZE - (len(buf) % IO_BLOCK_SIZE))

        return buf

    def add_file(self, filename, name=None, perms=None):
        if name is None:
            name = filename

        # If file exists, delete it
        for i in self.members:
            if i.name == name:
                # Remove from inode map
                if len(self.inodemap[i.ino]) > 1:
                    # Not going to deal with relocating data to another
                    # hard link
                    raise Exception("Trying to replace hard link with data")

                self.members.remove(i)
                self.inodemap[i.ino].remove(i)
                if len(self.inodemap[i.ino]) == 0:
                    del self.inodemap[i.ino]
                break

        member = CpioEntryNew()
        member.magic = MAGIC_NEW

        # Make sure we get some inode number if it's zero
        member.ino = os.stat(filename).st_ino
        if member.ino == 0:
            while True:
                member.ino -= 1
                if member.ino not in self.inodemap:
                    break

        if member.ino not in self.inodemap:
            # Make sure we only get the file contents once per inode
            self.inodemap[member.ino] = [ member ]

            with open(filename, 'rb') as f:
                member.content = f.read()
                member.filesize = len(member.content)

        else:
            self.inodemap[member.ino].append(member)

            member.content = b''
            member.filesize = 0

        fstat = os.stat(filename)

        if perms is not None:
            member.mode = stat.S_IFMT(fstat.st_mode)
            member.mode |= perms
        else:
            member.mode = os.stat(filename).st_mode

        member.uid = fstat.st_uid
        member.gid = fstat.st_gid

        if member.nlink == 0:
            if os.path.isdir(filename):
                member.nlink = 2
            else:
                # Assume hard links don't exist if Python can't get nlink value
                member.nlink = 1
        else:
            member.nlink = fstat.st_nlink

        member.mtime = os.path.getmtime(filename)
        if os.name == "nt":
            member.dev_maj = 0
            member.dev_min = 0
            member.rdev_maj = 0
            member.rdev_min = 0
        else:
            member.dev_maj = os.major(fstat.st_dev)
            member.dev_min = os.minor(fstat.st_dev)
            member.rdev_maj = os.major(fstat.st_rdev)
            member.rdev_min = os.minor(fstat.st_rdev)

        member.name = name
        member.namesize = len(name) + 1

        member.chksum = 0

        self.members.append(member)

    def get_file(self, name):
        for i in self.members:
            if i.name == name:
                return i

        return None


from droidtools import ext4fs

MODE_NORMAL = 0
MODE_SPARSED = 1
MODE_GZIP = 2

def make_ext4fs(filename, directory, length, mountPoint = None, mode = MODE_NORMAL):
    return ext4fs.make_ext4fs(filename, directory, mountPoint or "", str(int(length)), mode)


if __name__ == "__main__":
    make_ext4fs(
        "/home/tarek/.inception/out/suitepad/samsung-matissewifi/inselhof/cache.img",
        "/home/tarek/.inception/work/suitepad/samsung-matissewifi/inselhof/cache",
        314572800,
        "cache",
        MODE_SPARSED
        )




import tempfile
import unittest
import os
import shutil
from droidtools import ext4fs_utils

FS_SIZE     = 32 * 1024 * 1024
SPARSE_SIZE = 4771976
GZIP_SIZE   = 32917

class Ext4utilsTest(unittest.TestCase):
    def test_make(self):
        path = tempfile.mkdtemp()
        directory = os.path.join(path, "target")
        filepath = os.path.join(path, "target.img")
        os.makedirs(directory)
        ext4fs_utils.make_ext4fs(filepath, directory, FS_SIZE)
        self.assertEqual(os.path.getsize(filepath), FS_SIZE)
        shutil.rmtree(path)

    def test_sparsed(self):
        path = tempfile.mkdtemp()
        directory = os.path.join(path, "target")
        filepath = os.path.join(path, "target.img")
        os.makedirs(directory)
        ext4fs_utils.make_ext4fs(filepath, directory, FS_SIZE, mode = ext4fs_utils.MODE_SPARSED)
        self.assertEqual(os.path.getsize(filepath), SPARSE_SIZE)
        shutil.rmtree(path)

    def test_android(self):
        path = tempfile.mkdtemp()
        directory = os.path.join(path, "target")
        filepath = os.path.join(path, "target.img")
        os.makedirs(directory)
        ext4fs_utils.make_ext4fs(filepath, directory, FS_SIZE, mountPoint="cache")
        self.assertEqual(os.path.getsize(filepath), FS_SIZE)
        shutil.rmtree(path)

    def test_android_sparsed(self):
        path = tempfile.mkdtemp()
        directory = os.path.join(path, "target")
        filepath = os.path.join(path, "target.img")
        os.makedirs(directory)
        ext4fs_utils.make_ext4fs(filepath, directory, FS_SIZE, mountPoint="cache", mode=ext4fs_utils.MODE_SPARSED)
        self.assertEqual(os.path.getsize(filepath), SPARSE_SIZE)
        shutil.rmtree(path)

    def test_gzip(self):
        path = tempfile.mkdtemp()
        directory = os.path.join(path, "target")
        filepath = os.path.join(path, "target.img")
        os.makedirs(directory)
        ext4fs_utils.make_ext4fs(filepath, directory, FS_SIZE, mode=ext4fs_utils.MODE_GZIP)
        self.assertEqual(os.path.getsize(filepath), GZIP_SIZE)
        shutil.rmtree(path)